# -*- encoding: utf-8 -*-
from __future__ import print_function, unicode_literals, division, absolute_import
import logging
import socket

from enoceanx.communicators.communicator import Communicator


class TCPCommunicator(Communicator):
    ''' Socket communicator class for EnOcean radio '''
    logger = logging.getLogger('enocean.communicators.TCPCommunicator')

    def __init__(self, host='', port=9637):
        super(TCPCommunicator, self).__init__()
        self.host = host
        self.port = port

    def run(self):
        self.logger.info('TCPCommunicator started')
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind((self.host, self.port))
        sock.listen(5)
        sock.settimeout(0.5)

        while not self._stop_flag.is_set():
            try:
                (client, addr) = sock.accept()
            except socket.timeout:
                continue
            self.logger.debug('Client "%s" connected' % (addr))
            client.settimeout(0.5)
            while True and not self._stop_flag.is_set():
                try:
                    data = client.recv(2048)
                except socket.timeout:
                    break
                if not data:
                    break
                self._buffer.extend(bytearray(data))
            self.parse()
            client.close()
            self.logger.debug('Client disconnected')
        sock.close()
        self.logger.info('TCPCommunicator stopped')



# -*- encoding: utf-8 -*-
from __future__ import print_function, unicode_literals, division, absolute_import
from enum import IntEnum


# EnOceanSerialProtocol3.pdf / 12
class PACKET(IntEnum):
    RESERVED = 0x00
    # RADIO == RADIO_ERP1
    # Kept for backwards compatibility reasons, for example custom packet
    # generation shouldn't be affected...
    RADIO = 0x01
    RADIO_ERP1 = 0x01
    RESPONSE = 0x02
    RADIO_SUB_TEL = 0x03
    EVENT = 0x04
    COMMON_COMMAND = 0x05
    SMART_ACK_COMMAND = 0x06
    REMOTE_MAN_COMMAND = 0x07
    RADIO_MESSAGE = 0x09
    # RADIO_ADVANCED == RADIO_ERP2
    # Kept for backwards compatibility reasons
    RADIO_ADVANCED = 0x0A
    RADIO_ERP2 = 0x0A
    RADIO_802_15_4 = 0x10
    COMMAND_2_4 = 0x11


# EnOceanSerialProtocol3.pdf / 18
class RETURN_CODE(IntEnum):
    OK = 0x00
    ERROR = 0x01
    NOT_SUPPORTED = 0x02
    WRONG_PARAM = 0x03
    OPERATION_DENIED = 0x04


# EnOceanSerialProtocol3.pdf / 20
class EVENT_CODE(IntEnum):
    SA_RECLAIM_NOT_SUCCESFUL = 0x01
    SA_CONFIRM_LEARN = 0x02
    SA_LEARN_ACK = 0x03
    CO_READY = 0x04
    CO_EVENT_SECUREDEVICES = 0x05


# EnOcean_Equipment_Profiles_EEP_V2.61_public.pdf / 8
class RORG(IntEnum):
    UNDEFINED = 0x00
    RPS = 0xF6                      # Dec.: 246
    BS1 = 0xD5                      # Dec.: 213
    BS4 = 0xA5                      # Dec.: 165
    VLD = 0xD2                      # Dec.: 210
    MSC = 0xD1
    ADT = 0xA6
    SM_LRN_REQ = 0xC6
    SM_LRN_ANS = 0xC7
    SM_REC = 0xA7
    SYS_EX = 0xC5
    SEC = 0x30
    SEC_ENCAPS = 0x31
    UTE = 0xD4                      # Dec.: 212


# Results for message parsing
class PARSE_RESULT(IntEnum):
    OK = 0x00
    INCOMPLETE = 0x01
    CRC_MISMATCH = 0x03


# Data byte indexing
# Starts from the end, so works on messages of all length.
class DB0(object):
    BIT_0 = -1
    BIT_1 = -2
    BIT_2 = -3
    BIT_3 = -4
    BIT_4 = -5
    BIT_5 = -6
    BIT_6 = -7
    BIT_7 = -8


class DB1(object):
    BIT_0 = -9
    BIT_1 = -10
    BIT_2 = -11
    BIT_3 = -12
    BIT_4 = -13
    BIT_5 = -14
    BIT_6 = -15
    BIT_7 = -16


class DB2(object):
    BIT_0 = -17
    BIT_1 = -18
    BIT_2 = -19
    BIT_3 = -20
    BIT_4 = -21
    BIT_5 = -22
    BIT_6 = -23
    BIT_7 = -24


class DB3(object):
    BIT_0 = -25
    BIT_1 = -26
    BIT_2 = -27
    BIT_3 = -28
    BIT_4 = -29
    BIT_5 = -30
    BIT_6 = -31
    BIT_7 = -32


class DB4(object):
    BIT_0 = -33
    BIT_1 = -34
    BIT_2 = -35
    BIT_3 = -36
    BIT_4 = -37
    BIT_5 = -38
    BIT_6 = -39
    BIT_7 = -40


class DB5(object):
    BIT_0 = -41
    BIT_1 = -42
    BIT_2 = -43
    BIT_3 = -44
    BIT_4 = -45
    BIT_5 = -46
    BIT_6 = -47
    BIT_7 = -48


class DB6(object):
    BIT_0 = -49
    BIT_1 = -50
    BIT_2 = -51
    BIT_3 = -52
    BIT_4 = -53
    BIT_5 = -54
    BIT_6 = -55
    BIT_7 = -56



# -*- encoding: utf-8 -*-
from __future__ import print_function, unicode_literals, division, absolute_import
import logging


def get_bit(byte, bit):
    """ Get the bit value from byte """
    return (byte >> bit) & 0x01


def combine_hex(data):
    """ Combine list of integer values to one big integer """
    output = 0x00
    for i, value in enumerate(reversed(data)):
        output |= (value << i * 8)
    return output


def int_to_list(int_value):
    result = []
    while int_value > 0:
        result.append(int_value % 256)
        int_value = int_value // 256
    result.reverse()
    return result


def hex_to_list(hex_value):
    # it FFD97F81 has to be [FF, D9, 7F, 81] => [255, 217, 127, 129]
    result = []
    while hex_value > 0:
        result.append(hex_value % 0x100)
        hex_value = hex_value // 256
    result.reverse()
    return result


def to_bitarray(data, width=8):
    """ Convert data (list of integers, bytearray or integer) to bitarray """
    if isinstance(data, list) or isinstance(data, bytearray):
        data = combine_hex(data)
    return [True if digit == '1' else False for digit in bin(data)[2:].zfill(width)]


def from_bitarray(data):
    """ Convert bit array back to integer """
    return int(''.join(['1' if x else '0' for x in data]), 2)


def to_hex_string(data):
    if data is None:
        logging.getLogger().warning("cannot convert None to hex string")
        return 'NONE'
    ''' Convert list of integers to a hex string, separated by ":" '''
    if isinstance(data, int):
        return '%02X' % data
    return ':'.join([('%02X' % o) for o in data])


def from_hex_string(hex_string):
    reval = [int(x, 16) for x in hex_string.split(':')]
    if len(reval) == 1:
        return reval[0]
    return reval


def is_bs4_teach_in_packet(packet):
    """Checker whether it's a 4BS packet."""
    return len(packet.data) > 3 and get_bit(packet.data[4], 3) == 0


def get_next_free_base_id(start_base_id: int, used_base_ids_so_far: list):
    # print("Base ID given: %02x" % combine_hex(start_base_id))
    # used_base_ids_so_far is a list of lists (list of integer lists)
    # used_ids_as_str = ','.join([to_hex_string(base_id) for base_id in used_base_ids_so_far])
    # print("Used Base IDs so far: %s" % used_ids_as_str)

    communicator_max_base_id = combine_hex(start_base_id) + 127

    # print("Maximum Base ID: %s hex: " % str(communicator_max_base_id), to_hex_string(communicator_max_base_id))

    # is any Base ID used, yet?
    if len(used_base_ids_so_far) == 0:
        # use the start_base_id and return it
        base_id_as_list = int_to_list(start_base_id)
        used_base_ids_so_far.append(base_id_as_list)
        # return start_base_id
        return base_id_as_list

    # sort the ids
    # convert them to integers first
    sorted_used_base_ids_so_far = sorted([combine_hex(base_id) for base_id in used_base_ids_so_far])
    # used_base_ids_so_far.sort()
    # is there only one used (can't be zero here)
    if len(sorted_used_base_ids_so_far) == 1:
        # nothing to compare. Simply take the next one
        # return hex_to_list(sorted_used_base_ids_so_far[0] + 1)

        # check if the next one is smaller than the allowed limit and within the range of the base id
        possible_next_free_base_id = sorted_used_base_ids_so_far[0] + 1
        if possible_next_free_base_id <= 0xFFFFFFFE and possible_next_free_base_id <= communicator_max_base_id:
            return int_to_list(possible_next_free_base_id)

    for index, value in enumerate(sorted_used_base_ids_so_far[:-1]):
        # find a gap
        if sorted_used_base_ids_so_far[index+1] - sorted_used_base_ids_so_far[index] > 1:
            # there is a gap because the difference is bigger than one
            # so use the gap
            next_free_base_id = sorted_used_base_ids_so_far[index] + 1
            next_free_base_id_as_list = int_to_list(next_free_base_id)
            # this is only a copy: no appending & sorting required here
            # sorted_used_base_ids_so_far.append(next_free_base_id_as_list)
            # sorted_used_base_ids_so_far.sort()
            # convert them back to list of integers
            return next_free_base_id_as_list
        else:
            # there is no gap
            # so check the next one
            continue

    # we checked all the used Base IDs
    # now check if the next one is below the allowed limit of 0xFFFFFFFE
    # See (https://www.enocean-alliance.org/wp-content/uploads/2021/03/EURID-v1.2.pdf)
    # used_base_ids_so_far is a list of lists
    # last_used_base_id_as_list = used_base_ids_so_far[-1]
    last_used_base_id = sorted_used_base_ids_so_far[-1]
    next_free_base_id = last_used_base_id + 1
    if next_free_base_id <= 0xFFFFFFFE and next_free_base_id <= communicator_max_base_id:
        return int_to_list(next_free_base_id)
    else:
        # there is no Base ID left  # Exception?
        # return None
        raise OverflowError()



# -*- encoding: utf-8 -*-
from __future__ import print_function, unicode_literals, division, absolute_import
import logging
import serial
import time

from enoceanx.communicators.communicator import Communicator


class SerialCommunicator(Communicator):
    """ Serial port communicator class for EnOcean radio """
    logger = logging.getLogger('enocean.communicators.SerialCommunicator')

    # def __init__(self, port='/dev/ttyUSB0', callback=None):
    def __init__(self, port='/dev/ttyAMA0', callback=None):
        super(SerialCommunicator, self).__init__(callback)
        # Initialize serial port
        self.__ser = serial.Serial(port, 57600, timeout=0.1)

    def run(self):
        self.logger.info('SerialCommunicator started')
        while not self._stop_flag.is_set():
            # If there's messages in transmit queue
            # send them
            while True:
                packet = self._get_from_send_queue()
                if not packet:
                    break
                try:
                    self.__ser.write(bytearray(packet.build()))
                except serial.SerialException:
                    self.stop()

            # Read chars from serial port as hex numbers
            try:
                self._buffer.extend(bytearray(self.__ser.read(16)))
            except serial.SerialException:
                self.logger.error('Serial port exception! (device disconnected or multiple access on port?)')
                self.stop()
            self.parse()  # parse the buffer
            time.sleep(0)

        self.__ser.close()
        self.logger.info('SerialCommunicator stopped')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import time

from enoceanx.consolelogger import init_logging
import enoceanx.utils
from enoceanx.communicators.serialcommunicator import SerialCommunicator
from enoceanx.protocol.packet import RadioPacket
from enoceanx.protocol.constants import PACKET, RORG
import sys
import traceback

try:
    import queue
except ImportError:
    import Queue as queue


def assemble_radio_packet(transmitter_id):
    return RadioPacket.create(rorg=RORG.BS4, rorg_func=0x20, rorg_type=0x01,
                              sender=transmitter_id,
                              CV=50,
                              TMP=21.5,
                              ES='true')


init_logging()
communicator = SerialCommunicator(port="/dev/ttyUSB0")
communicator.start()
print('The Base ID of your module is %s.' % enoceanx.utils.to_hex_string(communicator.base_id))

# if communicator.base_id is not None:
    # print('Sending example package.')
    # communicator.send(assemble_radio_packet(communicator.base_id))

last_time = time.localtime()
time_in_seconds = time.time()
print("Time %s" % time.strftime("%H:%M", last_time))
# print(last_time.tm_sec)

# endless loop receiving radio packets
while communicator.is_alive():
    current_time = time.time()
    # if current_time > time.localtime(last_time.tm_sec):
    if current_time > time_in_seconds + 5:
        # print(last_time.tm_sec)
        time_in_seconds = current_time
        print("Communicator is still alive ...%s" % time.strftime("%H:%M:%S", time.localtime(current_time)))

    try:
        # Loop to empty the queue...
        packet = communicator.receive.get(block=True, timeout=1)
        if packet.packet_type == PACKET.RADIO_ERP1 and packet.rorg == RORG.VLD:
            packet.select_eep(0x01, 0x0e)  # check this
            packet.parse_eep()
            for k in packet.parsed:
                print('%s: %s' % (k, packet.parsed[k]))
        if packet.packet_type == PACKET.RADIO_ERP1 and packet.rorg == RORG.BS4:
            # parse packet with given FUNC and TYPE
            for k in packet.parse_eep(0x02, 0x05):
                print('%s: %s' % (k, packet.parsed[k]))
        if packet.packet_type == PACKET.RADIO_ERP1 and packet.rorg == RORG.BS1:
            # alternatively you can select FUNC and TYPE explicitely
            packet.select_eep(0x00, 0x01)                      # FUNC of window / door contact (FTKE): 00, TYPE: 01
            # parse it
            packet.parse_eep()
            for k in packet.parsed:
                print('%s: %s' % (k, packet.parsed[k]))
        if packet.packet_type == PACKET.RADIO_ERP1 and packet.rorg == RORG.RPS:
            # for k in packet.parse_eep(0x02, 0x02):
            for k in packet.parse_eep(0x10, 0x00):          # FUNC of window handle: 10, TYPE of window handle: 00
                print('%s: %s' % (k, packet.parsed[k]))
    except queue.Empty:
        continue
    except KeyboardInterrupt:
        break
    except Exception:
        traceback.print_exc(file=sys.stdout)
        break

if communicator.is_alive():
    communicator.stop()



from enum import IntEnum


class SmartAckCommandCode(IntEnum):
    SA_WR_LEARNMODE = 0x01  # Code 01 - Set/Reset Smart Ack learn mode
    SA_RD_LEARNMODE = 0x02  # Code 02 - Get Smart Ack learn mode state
    SA_WR_LEARNCONFIRM = 0x03  # Used for Smart Ack to add or delete a mailbox of a client
    SA_WR_CLIENTLEARNRQ = 0x04  # Send Smart Ack Learn request (Client)
    SA_WR_RESET = 0x05  # Send reset command to a Smart Ack client
    SA_RD_LEARNEDCLIENTS = 0x06  # Get Smart Ack learned sensors / mailboxes
    SA_WR_RECLAIMS = 0x07  # Set number of reclaim attempts
    SA_WR_POSTMASTER = 0x08  # Activate/Deactivate Post master functionality



''' Provider for different Communicator -classes for EnOcean. '''
from enoceanx.communicators.communicator import Communicator
from enoceanx.communicators.serialcommunicator import SerialCommunicator
from enoceanx.communicators.tcpcommunicator import TCPCommunicator



from enum import IntEnum


class CommonCommandCode(IntEnum):
    CO_WR_SLEEP = 0x1  # Code 01 - Enter energy saving mode
    CO_WR_RESET = 0x2  # Code 02 - Reset the device
    CO_RD_VERSION = 0x3  # Code 03 - Read the device version information
    CO_RD_SYS_LOG = 0x4  # Code 04 - Read system log
    CO_WR_SYS_LOG = 0x5  # Code 05 - Reset system log
    CO_WR_BIST = 0x6  # Code 06 - Perform Self Test
    CO_WR_IDBASE = 0x7  # Code 07 - Set ID range base address
    CO_RD_IDBASE = 0x8  # Code 08 - Read ID range base address
    CO_WR_REPEATER = 0x9  # Code 09 - Set Repeater Level
    CO_RD_REPEATER = 0xA  # Code 10 - Read Repeater Level
    CO_WR_FILTER_ADD = 0xB  # Code 11 - Add filter to list
    CO_WR_FILTER_DEL = 0xC  # Code 12 - Delete a specific filter from list
    CO_WR_FILTER_DEL_ALL = 0xD  # Code 13 - Delete all filters from filter list
    CO_WR_FILTER_ENABLE = 0xE  # Code 14 - Enable / Disable filter list
    CO_RD_FILTER = 0xF  # Code 15 - Read filters from filter list
    CO_WR_WAIT_MATURITY = 0x10  # Code 16 - Wait until the end of telegram maturity time before received radio telegrams
    # will be forwarded to the external host
    CO_WR_SUBTEL = 0x11  # Code 17 - Enable / Disable transmission of additional subtelegram info to the external host
    CO_WR_MEM = 0x12  # Code 18 - Write data to device memory
    CO_RD_MEM = 0x13  # Code 19 - Read data from device memory
    CO_RD_MEM_ADDRESS = 0x14  # Code 20 - Read address and length of the configuration area and the Smart Ack Table
    CO_RD_SECURITY = 0x15  # Code 21 -  DEPRECATED Read own security information (level, key)
    CO_WR_SECURITY = 0x16  # Code 22 - DEPRECATED Write own security information (level, key)
    CO_WR_LEARNMODE = 0x17  # Code 23 - Enable / disable learn mode
    CO_RD_LEARNMODE = 0x18  # Code 24 - Read learn mode status
    # TODO: up until Code 65



# -*- encoding: utf-8 -*-
from __future__ import print_function, unicode_literals, division, absolute_import
import logging
import datetime

import threading



try:
    import queue
except ImportError:
    import Queue as queue
from enoceanx.protocol.packet import Packet, UTETeachInPacket, RadioPacket, ResponsePacket
from enoceanx.protocol.constants import PACKET, PARSE_RESULT, RETURN_CODE, RORG
from enoceanx.protocol.common_command_codes import CommonCommandCode
from enoceanx import utils
# from enoceanx.protocol import teachin
# from enocean.protocol.teachin import TeachInHelper  # circular import -> teach in helper imports communicator


class Communicator(threading.Thread):
    """
    Communicator base-class for EnOcean.
    Not to be used directly, only serves as base class for SerialCommunicator etc.
    """
    logger = logging.getLogger('enocean.communicators.Communicator')

    def __init__(self, callback=None, teach_in=True):
        super(Communicator, self).__init__()
        # Create an event to stop the thread
        self._stop_flag = threading.Event()
        # Input buffer
        self._buffer = []
        # Setup packet queues
        self.transmit = queue.Queue()
        self.receive = queue.Queue()
        # Set the callback method
        self.__callback = callback
        # Internal variable for the Base ID of the module.
        self._base_id = None
        # Should new messages be learned automatically? Defaults to True.
        # Not sure if we should use CO_WR_LEARNMODE??
        # CO_WR_LEARNMODE is not supported by every device. The e.g. USB 300 does not support this command.

        # create an event to control whether teach-in is enabled (set (true) / clear (false))
        self._teach_in_flag = threading.Event()
        self.teach_in = teach_in

    def _get_from_send_queue(self):
        """ Get message from send queue, if one exists """
        try:
            packet = self.transmit.get(block=False)
            self.logger.info('Sending packet')
            self.logger.debug(packet)
            return packet
        except queue.Empty:
            pass
        return None

    def send(self, packet):
        if not isinstance(packet, Packet):
            self.logger.error('Object to send must be an instance of Packet')
            return False
        self.transmit.put(packet)
        return True

    def stop(self):
        self._stop_flag.set()

    def deactivate_teach_in(self):
        self.logger.info("Teach-in DISABLED")
        self._teach_in_flag.clear()

    def activate_teach_in(self):
        self.logger.info("Teach-in ENABLED")
        self._teach_in_flag.set()

    def parse(self):
        """ Parses messages and puts them to receive queue """
        # Loop while we get new messages
        while True:
            status, self._buffer, packet = Packet.parse_msg(self._buffer)
            # If message is incomplete -> break the loop
            if status == PARSE_RESULT.INCOMPLETE:
                return status

            # If message is OK, add it to receive queue or send to the callback method
            if status == PARSE_RESULT.OK and packet:
                packet.received = datetime.datetime.now()

                if isinstance(packet, UTETeachInPacket) and self.teach_in:
                    response_packet = packet.create_response_packet(self.base_id)
                    self.logger.info('Sending response to UTE teach-in.')
                    self.send(response_packet)

                # TODO: teach-in handling for other RORGS
                if isinstance(packet, RadioPacket) \
                        and packet.packet_type == PACKET.RADIO_ERP1 \
                        and packet.rorg == RORG.BS4:
                    # check for teach in packet AND
                    # TODO: refactor this
                    # for BS4 the third bit of the fourth byte determines if the packet is a teach-in packet
                    if self._teach_in_flag is True and utils.get_bit(packet.data[4], 3) == 0:
                        # if self.teach_in is True and utils.get_bit(packet.data[4], 3) == 0:
                        # we have a BS4 teach-in packet AND we want to teach-in the new device
                        # remove print statements
                        # TODO: extract necessary and optional data like eep, sender_id
                        # packet.sender
                        # if packet.contains_eep:
                        teach_in_response_packet = TeachInHelper.create_bs4_teach_in_response(packet, self)

                        self.logger.info("BS4 teach-in response created")

                        # send the packet
                        # TODO: send here already? What about callback function
                        successful_sent = self.send(teach_in_response_packet)

                        if successful_sent:
                            # the package was put to the transmit queue
                            self.logger.info("Sent teach-in response")
                            # print("Sent teach-in response")
                        # TODO: here

                if self.__callback is None:
                    self.receive.put(packet)
                else:
                    # TODO: maybe extract data in callback
                    self.__callback(packet)
                self.logger.debug(packet)

    @property  # getter
    def callback(self):
        return self.__callback

    @callback.setter
    def callback(self, callback):
        self.__callback = callback

    @property
    def base_id(self):
        """ Fetches Base ID from the transmitter, if required, otherwise returns the currently set Base ID. """
        # If base id is already set, return it.
        if self._base_id is not None:
            return self._base_id

        # Send COMMON_COMMAND 0x08, CO_RD_IDBASE request to the module
        # self.send(Packet(PACKET.COMMON_COMMAND, data=[0x08]))  # the next line does the same
        self.send(Packet(PACKET.COMMON_COMMAND, data=[CommonCommandCode.CO_RD_IDBASE]))
        # Loop over 10 times, to make sure we catch the response.
        # Thanks to timeout, shouldn't take more than a second.
        # Unfortunately, all other messages received during this time are ignored.
        for i in range(0, 10):
            try:
                packet: ResponsePacket = self.receive.get(block=True, timeout=0.1)
                # We're only interested in responses to the request in question.
                if packet.packet_type == PACKET.RESPONSE and packet.response == RETURN_CODE.OK and len(packet.response_data) == 4:  # noqa: E501
                    # Base ID is set in the response data.
                    self._base_id = packet.response_data
                    # Put packet back to the Queue, so the user can also react to it if required...
                    self.receive.put(packet)
                    break
                # Put other packets back to the Queue.
                self.receive.put(packet)
            except queue.Empty:
                continue
        # Return the current Base ID (might be None).
        return self._base_id

    @base_id.setter
    def base_id(self, base_id):
        """ Sets the Base ID manually, only for testing purposes. """
        self._base_id = base_id


class TeachInHelper:
    logger = logging.getLogger()

    @staticmethod
    def create_bs4_teach_in_response(cls, packet: Packet, communicator: Communicator) -> RadioPacket:
        # let's create a proper response
        rorg = packet.rorg
        # print("rorg of requesting device: %s" % str(rorg))
        cls.logger.info("rorg of requesting device: %s" % str(rorg))
        func = packet.rorg_func
        cls.logger.info("rorg_func of requesting device: %s" % str(func))
        # print("rorg_func of requesting device: %s" % str(func))
        rorg_type = packet.rorg_type
        cls.logger.info("rorg_type of requesting device: %s" % str(rorg_type))
        # print("rorg_type of requesting device: %s" % str(rorg_type))
        base_id = communicator.base_id

        teach_in_response_packet: RadioPacket = \
            Packet.create(PACKET.RADIO_ERP1,
                          # rorg=RORG.BS4,  # respond with BS4 teach-in-response
                          rorg=rorg,
                          rorg_func=func,
                          rorg_type=rorg_type,
                          sender=base_id,
                          learn=True)
        # copy over the packet data as it will be sent back with slight variation
        teach_in_response_packet.data[1:5] = packet.data[1:5]

        # set the bits of the byte for the success case (F0 = 11110000)
        teach_in_response_packet.data[4] = 0xF0

        # teach_in_response_packet.destination = packet.
        # set destination to former sender
        destination = packet.data[-5:-1]
        teach_in_response_packet.destination = destination

        # set sender to base id (no offset)
        # TODO: test base id + 1
        # maybe use utils to get the next free base_id
        # base_id = self.base_id
        cls.logger.info("base id: {}" % base_id)
        # print("base id: {}" % base_id)
        teach_in_response_packet.sender = base_id

        # build the optional data
        # subTelegram Number + destination + dBm (send case: FF) + security (0)
        optional = [3] + destination + [0xFF, 0]
        teach_in_response_packet.optional = optional
        radio_response_packet = teach_in_response_packet.parse()
        cls.logger.info("BS4 teach-in response created")

        return radio_response_packet



import enoceanx

import logging

from build.lib.enoceanx.protocol.packet import RadioPacket
from enoceanx.communicators import Communicator
from enoceanx.protocol.constants import PACKET
from enoceanx.protocol.packet import Packet


class TeachInHelper:
    logger: logging.Logger = logging.getLogger()

    @staticmethod
    def create_bs4_teach_in_response(cls, packet: Packet, communicator: Communicator) -> RadioPacket:
        # let's create a proper response
        rorg = packet.rorg
        # print("rorg of requesting device: %s" % str(rorg))
        cls.logger.info("rorg of requesting device: %s" % str(rorg))
        func = packet.rorg_func
        cls.logger.info("rorg_func of requesting device: %s" % str(func))
        # print("rorg_func of requesting device: %s" % str(func))
        rorg_type = packet.rorg_type
        cls.logger.info("rorg_type of requesting device: %s" % str(rorg_type))
        # print("rorg_type of requesting device: %s" % str(rorg_type))
        base_id = communicator.base_id

        teach_in_response_packet: RadioPacket = \
            Packet.create(PACKET.RADIO_ERP1,
                          # rorg=RORG.BS4,  # respond with BS4 teach-in-response
                          rorg=rorg,
                          rorg_func=func,
                          rorg_type=rorg_type,
                          sender=base_id,
                          learn=True)
        # copy over the packet data as it will be sent back with slight variation
        teach_in_response_packet.data[1:5] = packet.data[1:5]

        # set the bits of the byte for the success case (F0 = 11110000)
        teach_in_response_packet.data[4] = 0xF0

        # teach_in_response_packet.destination = packet.
        # set destination to former sender
        destination = packet.data[-5:-1]
        teach_in_response_packet.destination = destination

        # set sender to base id (no offset)
        # TODO: test base id + 1
        # maybe use utils to get the next free base_id
        # base_id = self.base_id
        cls.logger.debug("base id: {}" % base_id)
        # print("base id: {}" % base_id)
        teach_in_response_packet.sender = base_id

        # build the optional data
        # subTelegram Number + destination + dBm (send case: FF) + security (0)
        optional = [3] + destination + [0xFF, 0]
        teach_in_response_packet.optional = optional
        radio_response_packet = teach_in_response_packet.parse()
        cls.logger.debug("BS4 teach-in response created")

        return radio_response_packet




#!/usr/bin/env python
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='enoceanx',
    version='0.63',
    description='EnOcean serial protocol implementation with support for 4BS teach-in. This is just a fork of enocean by kipe. Credits belong to the original author.',
    author='Felix Beier',
    author_email='choleriker99@gmail.com',
    url='https://github.com/flexxor/enoceanx',
    packages=[
        'enoceanx',
        'enoceanx.protocol',
        'enoceanx.communicators',
    ],
    scripts=[
        'examples/enocean_example.py',
    ],
    package_data={
        '': ['EEP.xml']
    },
    install_requires=[
        'enum-compat>=0.0.2',
        'pyserial>=3.0',
        'beautifulsoup4>=4.3.2',
    ])
